;;;; crunch - primitive declarations

(module (crunch declarations)

  (define-overloaded-primitive
   define-inline-primitive
   define-custom-inline-primitive
   define-primitive
   define-specialization
   :)

  (import (scheme base))
  (import (chicken plist))
  (import matchable)

  (import (crunch compiler type))

  (import-for-syntax (chicken plist) matchable)
  (import-for-syntax (crunch compiler type))

  (define-syntax :
    (er-macro-transformer
      (lambda (x r c)
        (define (decl n t safe)
          (let ((a (##sys#alias-global-hook n #t #f)))
            (put! a 'crunch:declared-type (typespec t))
            (when safe (put! a 'crunch:safe #t))))
        (define (m2 a result safe)
          (match a
            ((name args ...) (decl name `(procedure ,args ,result) safe))
            ((? symbol? name) (decl name result safe))
            (_ (error ': "ill-formed declaration" x))))
        (match (strip-syntax x)
          ((_ 'safe a b) (m2 a b #t))
          ((_ a b) (m2 a b #f))
          (_ (error ': "ill-formed declaration" x)))
        '(##core#undefined))))

  (define-syntax define-overloaded-primitive
    (syntax-rules ()
      ((_ (name args ...) rtype id rules ...)
        (begin
          (define-primitive (name args ...) rtype id)
          (put! 'id 'crunch:overload
            (match-lambda rules ... (_ #f)))))))

  (define-syntax define-primitive
    (syntax-rules ()
      ((_ (name args ...) rtype id)
        (begin
          (put! 'name 'crunch:function-id 'id)
          (put! 'id 'crunch:primitive 'name)
          (put! 'name 'crunch:primitive #t)
          (put! 'name 'crunch:type
                       (typespec '(procedure (args ...) rtype) 'id))))))

  (define-syntax define-inline-primitive
    (syntax-rules ()
      ((_ (name args ...) expr)
        (put! 'name 'crunch:inline '(lambda (args ...) expr)))))

  (define-syntax define-custom-inline-primitive
    (syntax-rules ()
      ((_ name proc)
        (put! 'name 'crunch:inline proc))))

  (define-syntax define-specialization
    (syntax-rules (=)
      ((_ id = other)
        (put! 'id 'crunch:specialize (get 'other 'crunch:specialize)))
      ((_ id rules ...)
        (put! 'id 'crunch:specialize
          (match-lambda rules ... (_ #f))))))

)
